package gs.test.v0.definition

import cats.Applicative
import cats.data.EitherT
import cats.effect.Sync
import cats.syntax.all.*
import gs.test.v0.definition.pos.SourcePosition

/** Type alias for test results. Test results are either a [[TestFailure]] or a
  * unit value (indicating success).
  */
type TestResult = Either[TestFailure, Unit]

/** String interpolator for [[Tag]]. Shorthand for producing new [[Tag]]
  * instances.
  *
  * {{{
  * import gs.test.v0.definition.*
  * val tag1: TestDefinition.Tag = tag"example"
  * }}}
  */
extension (sc: StringContext) def tag(args: Any*): Tag = Tag(sc.s(args*))

/** String interpolator for [[PermanentId]]. Shorthand for producing new
  * [[PermanentId]] instances.
  *
  * {{{
  * import gs.test.v0.definition.*
  * val permanentId: PermanentId = pid"example"
  * }}}
  */
extension (sc: StringContext)
  def pid(args: Any*): PermanentId = PermanentId(sc.s(args*))

/** Request this test to fail (pure form).
  *
  * @param message
  *   The message to report - why did this test fail?
  * @return
  *   The failing test result.
  */
def fail(
  message: String
)(
  using
  pos: SourcePosition
): Either[TestFailure, Unit] =
  Left(TestFailure.TestRequestedFailure(message, pos))

/** Request this test to fail (lifted into F).
  *
  * @param message
  *   The message to report - why did this test fail?
  * @return
  *   The failing test result.
  */
def failF[F[_]: Applicative](
  message: String
)(
  using
  pos: SourcePosition
): F[Either[TestFailure, Unit]] =
  Applicative[F].pure(fail(message))

/** Request this test to fail (lifted into EitherT).
  *
  * @param message
  *   The message to report - why did this test fail?
  * @return
  *   The failing test result.
  */
def failT[F[_]: Applicative](
  message: String
)(
  using
  pos: SourcePosition
): EitherT[F, TestFailure, Unit] =
  EitherT(failF(message))

/** Shorthand for indicating a passing test (pure form).
  *
  * ## Example
  *
  * {{{
  * import gs.test.v0.definition.*
  *
  * final class Example extends TestGroup.IO:
  *   override def name: String = "example"
  *
  *   test(pid"ex", "Example Test").pure { pass() }
  * }}}
  *
  * @return
  *   The passing test result.
  */
def pass(): Either[TestFailure, Unit] = Right(())

/** Shorthand for indicating a passing test (lifted into F).
  *
  * ## Example
  *
  * {{{
  * import gs.test.v0.definition.*
  *
  * final class Example extends TestGroup.IO:
  *   override def name: String = "example"
  *
  *   test(pid"ex", "Example Test").effectful { passF() }
  * }}}
  *
  * @return
  *   The passing test result.
  */
def passF[F[_]: Applicative](): F[Either[TestFailure, Unit]] =
  Applicative[F].pure(Right(()))

/** Shorthand for indicating a passing test (lifted into EitherT).
  *
  * ## Example
  *
  * {{{
  * import gs.test.v0.definition.*
  *
  * final class Example extends TestGroup.IO:
  *   override def name: String = "example"
  *
  *   test(pid"ex", "Example Test") { passT() }
  * }}}
  *
  * @return
  *   The passing test result.
  */
def passT[F[_]: Applicative](): EitherT[F, TestFailure, Unit] =
  EitherT(passF())

/** Check all of the given results, returning the first failure, or a successful
  * result if no result failed.
  *
  * @param results
  *   The list of results to check.
  * @return
  *   Successful result or the first failure.
  */
def checkAll(
  results: Either[TestFailure, Unit]*
): Either[TestFailure, Unit] =
  val initial: Either[TestFailure, Unit] = Right(())
  results.foldLeft(initial) {
    (
      acc,
      result
    ) =>
      acc match
        case Left(_)  => acc
        case Right(_) => result
  }

/** Check all of the given results, returning the first failure, or a successful
  * result if no result failed.
  *
  * In this call, each check is expressed as an effect that is evaluated to
  * determine success/failure - this call is lazy.
  *
  * @param results
  *   The list of results to check.
  * @return
  *   Successful result or the first failure.
  */
def checkAllF[F[_]: Sync](
  checks: F[Either[TestFailure, Unit]]*
): F[Either[TestFailure, Unit]] =
  val initial: F[Either[TestFailure, Unit]] = Sync[F].delay(Right(()))
  checks.foldLeft(initial) {
    (
      acc,
      result
    ) =>
      acc.flatMap {
        case Right(_) => result
        case err      => Sync[F].pure(err)
      }
  }

def check[A](candidate: A): Check[A] = Check(candidate)
